@@ -5,8 +5,13 @@
 
 pub use crate::context::{Context, ContextFlags};
 pub use crate::device::Device;
+pub use crate::event::{Event, EventFlags, EventStatus};
+pub use crate::external::*;
+pub use crate::function::Function;
 pub use crate::launch;
-pub use crate::memory::{CopyDestination, DeviceBuffer, UnifiedBuffer};
+pub use crate::memory::{
+    CopyDestination, DeviceBuffer, DevicePointer, DeviceSlice, DeviceVariable, UnifiedBuffer,
+};
 pub use crate::module::Module;
 pub use crate::stream::{Stream, StreamFlags};
 pub use crate::util::*;
